/*
 * Copyright (c) 2016, NVIDIA CORPORATION. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *  * Neither the name of NVIDIA CORPORATION nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "./include/NvUtils.h"
#include <iostream>
#include <linux/videodev2.h>
#include <malloc.h>
#include <string.h>
#include <fcntl.h>
#include <poll.h>
#include <unistd.h>
#include <sys/prctl.h>
#include "./include/nvbuf_utils.h"
#include "nvenc_main.h"
#include "video_encode.h"

#define TEST_ERROR(cond, str, label) if(cond) { \
                                        LOGE(str"\n"); \
                                        error = 1; \
                                        goto label; }

#define TEST_PARSE_ERROR(cond, label) if(cond) { \
    LOGE("Error parsing runtime parameter changes string"); \
    goto label; }

#define IS_DIGIT(c) (c >= '0' && c <= '9')
#define MICROSECOND_UNIT 1000000

static void
abort(context_t *ctx)
{
    ctx->got_error = true;
    ctx->enc->abort();
}


static int
write_encoder_output_frame(NvBuffer * buffer, int (* funcWriteBuf)(unsigned char *, uint32_t))
{
    if (NULL != funcWriteBuf)
    {
        funcWriteBuf(buffer->planes[0].data, buffer->planes[0].bytesused);
    }
    return 0;
}

static bool
encoder_capture_plane_dq_callback(struct v4l2_buffer *v4l2_buf, NvBuffer * buffer,
                                  NvBuffer * shared_buffer, void *arg)
{
    if (NULL == buffer)
    {
        LOGE("input param error!\n");
        return false;
    }
    if (NULL == arg)
    {
        LOGE("input param error!\n");
        return false;
    }
    context_t *ctx = (context_t *) arg;
    NvVideoEncoder *enc = ctx->enc;
    uint32_t frame_num = ctx->enc->capture_plane.getTotalDequeuedBuffers() - 1;

    if (v4l2_buf == NULL)
    {
        LOGE("Error while dequeing buffer from output plane\n");
        abort(ctx);
        return false;
    }

    // GOT EOS from encoder. Stop dqthread.
    if (buffer->planes[0].bytesused == 0)
    {
        LOGE("Got 0 size buffer in capture.\n");
        return false;
    }
    if (NULL != ctx->funcWriteBuf)
    {
        write_encoder_output_frame(buffer, ctx->funcWriteBuf);    /**> for writing file */
    }

    if (enc->capture_plane.qBuffer(*v4l2_buf, NULL) < 0)
    {
        LOGE("Error while Qing buffer at capture plane\n");
        abort(ctx);
        return false;
    }

    return true;
}

static void
set_defaults(context_t * ctx)
{
    memset(ctx, 0, sizeof(context_t));

    ctx->bitrate = 200*1024;
    ctx->peak_bitrate = 0;
    ctx->profile = V4L2_MPEG_VIDEO_H264_PROFILE_HIGH;
    ctx->ratecontrol = V4L2_MPEG_VIDEO_BITRATE_MODE_CBR;
    ctx->iframe_interval = 15;
    ctx->nH264FrameNumBits = 0;
    ctx->nH265PocLsbBits = 0;
    ctx->idr_interval = 15;
    ctx->level = V4L2_MPEG_VIDEO_H264_LEVEL_5_1;
    ctx->fps_n = 15;
    ctx->fps_d = 1;
    ctx->gdr_start_frame_number = 0xffffffff;
    ctx->gdr_num_frames = 0xffffffff;
    ctx->gdr_out_frame_number = 0xffffffff;
    ctx->num_b_frames = (uint32_t) -1;
    ctx->nMinQpI = (uint32_t)QP_RETAIN_VAL;
    ctx->nMaxQpI = (uint32_t)QP_RETAIN_VAL;
    ctx->nMinQpP = (uint32_t)QP_RETAIN_VAL;
    ctx->nMaxQpP = (uint32_t)QP_RETAIN_VAL;
    ctx->nMinQpB = (uint32_t)QP_RETAIN_VAL;
    ctx->nMaxQpB = (uint32_t)QP_RETAIN_VAL;
    ctx->sMaxQp = 51;
    ctx->output_memory_type = V4L2_MEMORY_DMABUF;
    ctx->copy_timestamp = true;
    ctx->start_ts = 0;
    ctx->blocking_mode = 1;
    ctx->hw_preset_type = V4L2_ENC_HW_PRESET_ULTRAFAST;
}


static int
setup_output_dmabuf(context_t *ctx, uint32_t num_buffers )
{
    int ret=0;
    NvBufferCreateParams cParams;
    int fd;
    ret = ctx->enc->output_plane.reqbufs(V4L2_MEMORY_DMABUF,num_buffers);
    if(ret)
    {
        LOGE("reqbufs failed for output plane V4L2_MEMORY_DMABUF\n");
        return ret;
    }
    for (uint32_t i = 0; i < ctx->enc->output_plane.getNumBuffers(); i++)
    {
        cParams.width = ctx->width;
        cParams.height = ctx->height;
        cParams.layout = NvBufferLayout_Pitch;
        cParams.colorFormat = NvBufferColorFormat_YUV420;
        cParams.nvbuf_tag = NvBufferTag_VIDEO_ENC;
        cParams.payloadType = NvBufferPayload_SurfArray;
        ret = NvBufferCreateEx(&fd, &cParams);
        if(ret < 0)
        {
            LOGE("Failed to create NvBuffer\n");
            return ret;
        }
        ctx->output_plane_fd[i]=fd;
    }
    return ret;
}
extern int g_bitrate;
static int encoder_proc_blocking(context_t &ctx, bool eos)
{
    int ret = 0;
    // Keep reading input till EOS is reached
    while (!ctx.got_error && !ctx.enc->isInError() && !eos)
    {
        if (ctx.bitrate != g_bitrate * 1024)
        {
            LOGI("encoder bitrate changed! exit this thread. now bitrate: %d, last bitrate: %d\n", \
                ctx.bitrate, g_bitrate * 1024);
            goto cleanup;
        }
        struct v4l2_buffer v4l2_buf;
        struct v4l2_plane planes[MAX_PLANES];
        NvBuffer *buffer;

        memset(&v4l2_buf, 0, sizeof(v4l2_buf));
        memset(planes, 0, sizeof(planes));

        v4l2_buf.m.planes = planes;

        if (ctx.enc->output_plane.dqBuffer(v4l2_buf, &buffer, NULL, 10) < 0)
        {
            LOGE("ERROR while DQing buffer at output plane\n");
            abort(&ctx);
            goto cleanup;
        }
        if (read_video_frame(ctx.in_file, ctx.read_camera, ctx.enc_sem, *buffer) < 0)
        {
            LOGE("Could not read complete frame from input file\n");
            v4l2_buf.m.planes[0].bytesused = 0;            
            eos = true;
            ctx.got_eos = true;
            v4l2_buf.m.planes[0].m.userptr = 0;
            v4l2_buf.m.planes[0].bytesused = v4l2_buf.m.planes[1].bytesused = v4l2_buf.m.planes[2].bytesused = 0;
        }

        if (ctx.copy_timestamp)
        {
          v4l2_buf.flags |= V4L2_BUF_FLAG_TIMESTAMP_COPY;
          ctx.timestamp += ctx.timestampincr;
          v4l2_buf.timestamp.tv_sec = ctx.timestamp / (MICROSECOND_UNIT);
          v4l2_buf.timestamp.tv_usec = ctx.timestamp % (MICROSECOND_UNIT);
        }

        if(ctx.output_memory_type == V4L2_MEMORY_DMABUF || ctx.output_memory_type == V4L2_MEMORY_MMAP)
        {
            for (int j = 0 ; j < MAX_PLANES ; j++)
            {
                ret = NvBufferMemSyncForDevice (buffer->planes[j].fd, j, (void **)&buffer->planes[j].data);
                if (ret < 0)
                {
                    LOGE("Error while NvBufferMemSyncForDevice at output plane for V4L2_MEMORY_DMABUF\n");
                    abort(&ctx);
                    goto cleanup;
                }
            }
        }

        if(ctx.output_memory_type == V4L2_MEMORY_DMABUF)
        {
            for (int j = 0 ; j < MAX_PLANES ; j++)
            {
                v4l2_buf.m.planes[j].bytesused = buffer->planes[j].bytesused;
            }
        }
        ret = ctx.enc->output_plane.qBuffer(v4l2_buf, NULL);
        if (ret < 0)
        {
            LOGE("Error while queueing buffer at output plane\n");
            abort(&ctx);
            goto cleanup;
        }

        ctx.input_frames_queued_count++;
        if (v4l2_buf.m.planes[0].bytesused == 0)
        {
            LOGI("File read complete.\n");
            eos = true;
            ctx.got_eos = true;
            return 0;
        }
    }
cleanup:
    return -1;
}


void *nvenc_main(void *arg)
{
    prctl(PR_SET_NAME, __FUNCTION__);
    THREAD_PARAM *threadParam;
    threadParam = (THREAD_PARAM *)arg;
    LOGI("fileSrc=%p, width=%d, heigth=%d, bitrate=%d, fps=%d\n", \
        threadParam->fileSrc, threadParam->width, threadParam->heigth, threadParam->bitrate, threadParam->fps);
    context_t ctx = {0};
    int ret = 0;
    int error = 0;
    bool eos = false;
    set_defaults(&ctx);
    ctx.in_file = threadParam->fileSrc;
    ctx.width = 640;
    ctx.height = 480;
    ctx.encoder_pixfmt = V4L2_PIX_FMT_H264;
    if (2 == threadParam->encType)
    {
        ctx.encoder_pixfmt = V4L2_PIX_FMT_H265;
    }
    ctx.bitrate = threadParam->bitrate;
    ctx.read_camera = threadParam->read_camera;
    ctx.enc_sem = threadParam->enc_sem;
    ctx.funcWriteBuf = threadParam->funcWriteBuf;

    if (ctx.encoder_pixfmt == V4L2_PIX_FMT_H265)
    {
        TEST_ERROR(ctx.width < 144 || ctx.height < 144, "Height/Width should be > 144 for H.265", cleanup);
    }

    if (ctx.blocking_mode)
    {
        LOGD("Creating Encoder in blocking mode.\n");
        ctx.enc = NvVideoEncoder::createVideoEncoder("enc0");
    }
    TEST_ERROR(NULL == ctx.enc, "Could not create encoder", cleanup);

    // It is necessary that Capture Plane format be set before Output Plane
    // format.
    // Set encoder capture plane format. It is necessary to set width and
    // height on thr capture plane as well
    ret =
        ctx.enc->setCapturePlaneFormat(ctx.encoder_pixfmt, ctx.width,
                                      ctx.height, 2 * 1024 * 1024);
    TEST_ERROR(ret < 0, "Could not set output plane format", cleanup);

    // Set encoder output plane format
    ret =
        ctx.enc->setOutputPlaneFormat(V4L2_PIX_FMT_YUV420M, ctx.width,
                                      ctx.height);
    TEST_ERROR(ret < 0, "Could not set output plane format", cleanup);

    ret = ctx.enc->setBitrate(ctx.bitrate);
    TEST_ERROR(ret < 0, "Could not set encoder bitrate", cleanup);

    ret = ctx.enc->setProfile(ctx.profile);
    TEST_ERROR(ret < 0, "Could not set encoder profile", cleanup);

    if (ctx.encoder_pixfmt == V4L2_PIX_FMT_H264)
    {
        ret = ctx.enc->setLevel(ctx.level);
        TEST_ERROR(ret < 0, "Could not set encoder level", cleanup);
    }

    ret = ctx.enc->setIDRInterval(ctx.idr_interval);
    TEST_ERROR(ret < 0, "Could not set encoder IDR interval", cleanup);

    ret = ctx.enc->setInsertSpsPpsAtIdrEnabled(true);
    TEST_ERROR(ret < 0, "Could not set insertSPSPPSAtIDR", cleanup);

    ret = ctx.enc->setIFrameInterval(ctx.iframe_interval);
    TEST_ERROR(ret < 0, "Could not set encoder I-Frame interval", cleanup);

    ret = ctx.enc->setFrameRate(ctx.fps_n, ctx.fps_d);
    TEST_ERROR(ret < 0, "Could not set framerate", cleanup);

    if (ctx.hw_preset_type)
    {
        ret = ctx.enc->setHWPresetType(ctx.hw_preset_type);
        TEST_ERROR(ret < 0, "Could not set encoder HW Preset Type", cleanup);
    }

    if (ctx.num_b_frames != (uint32_t) -1)
    {
        ret = ctx.enc->setNumBFrames(ctx.num_b_frames);
        TEST_ERROR(ret < 0, "Could not set number of B Frames", cleanup);
    }

    if ((ctx.nMinQpI != (uint32_t)QP_RETAIN_VAL) ||
        (ctx.nMaxQpI != (uint32_t)QP_RETAIN_VAL) ||
        (ctx.nMinQpP != (uint32_t)QP_RETAIN_VAL) ||
        (ctx.nMaxQpP != (uint32_t)QP_RETAIN_VAL) ||
        (ctx.nMinQpB != (uint32_t)QP_RETAIN_VAL) ||
        (ctx.nMaxQpB != (uint32_t)QP_RETAIN_VAL))
    {
        ret = ctx.enc->setQpRange(ctx.nMinQpI, ctx.nMaxQpI, ctx.nMinQpP,
                ctx.nMaxQpP, ctx.nMinQpB, ctx.nMaxQpB);
        TEST_ERROR(ret < 0, "Could not set quantization parameters", cleanup);
    }

    // Query, Export and Map the output plane buffers so that we can read
    // raw data into the buffers
    ret = setup_output_dmabuf(&ctx,10);
    TEST_ERROR(ret < 0, "Could not setup plane", cleanup);

    ret = ctx.enc->capture_plane.setupPlane(V4L2_MEMORY_MMAP, 10, true, false);
    TEST_ERROR(ret < 0, "Could not setup capture plane", cleanup);

    ret = ctx.enc->subscribeEvent(V4L2_EVENT_EOS,0,0);
    TEST_ERROR(ret < 0, "Could not subscribe EOS event", cleanup);

    // output plane STREAMON
    ret = ctx.enc->output_plane.setStreamStatus(true);
    TEST_ERROR(ret < 0, "Error in output plane streamon", cleanup);

    // capture plane STREAMON
    ret = ctx.enc->capture_plane.setStreamStatus(true);
    TEST_ERROR(ret < 0, "Error in capture plane streamon", cleanup);


    if (ctx.blocking_mode)
    {
        ctx.enc->capture_plane.
            setDQThreadCallback(encoder_capture_plane_dq_callback);

        // startDQThread starts a thread internally which calls the
        // encoder_capture_plane_dq_callback whenever a buffer is dequeued
        // on the plane
        ctx.enc->capture_plane.startDQThread(&ctx);
    }

    // Enqueue all the empty capture plane buffers
    for (uint32_t i = 0; i < ctx.enc->capture_plane.getNumBuffers(); i++)
    {
        struct v4l2_buffer v4l2_buf;
        struct v4l2_plane planes[MAX_PLANES];

        memset(&v4l2_buf, 0, sizeof(v4l2_buf));
        memset(planes, 0, MAX_PLANES * sizeof(struct v4l2_plane));

        v4l2_buf.index = i;
        v4l2_buf.m.planes = planes;

        ret = ctx.enc->capture_plane.qBuffer(v4l2_buf, NULL);
        if (ret < 0)
        {
            LOGE("Error while queueing buffer at capture plane\n");
            abort(&ctx);
            goto cleanup;
        }
    }

    if (ctx.copy_timestamp)
    {
      ctx.timestamp = (ctx.start_ts * MICROSECOND_UNIT);
      ctx.timestampincr = (MICROSECOND_UNIT * 16) / ((uint32_t) (ctx.fps_n * 16));
    }

    // Read video frame and queue all the output plane buffers
    for (uint32_t i = 0; i < ctx.enc->output_plane.getNumBuffers(); i++)
    {
        struct v4l2_buffer v4l2_buf;
        struct v4l2_plane planes[MAX_PLANES];
        NvBuffer *buffer = ctx.enc->output_plane.getNthBuffer(i);

        memset(&v4l2_buf, 0, sizeof(v4l2_buf));
        memset(planes, 0, MAX_PLANES * sizeof(struct v4l2_plane));

        v4l2_buf.index = i;
        v4l2_buf.m.planes = planes;

        if(ctx.output_memory_type == V4L2_MEMORY_DMABUF)
        {
            v4l2_buf.type = V4L2_BUF_TYPE_VIDEO_OUTPUT;
            v4l2_buf.memory = V4L2_MEMORY_DMABUF;
            ret = ctx.enc->output_plane.mapOutputBuffers(v4l2_buf, ctx.output_plane_fd[i]);

            if (ret < 0)
            {
                LOGE("Error while mapping buffer at output plane\n");
                abort(&ctx);
                goto cleanup;
            }
        }
        if (read_video_frame(ctx.in_file, ctx.read_camera, NULL, *buffer) < 0)
        {
            LOGE("Could not read complete frame from input file\n");
            v4l2_buf.m.planes[0].bytesused = 0;
            eos = true;
            v4l2_buf.m.planes[0].m.userptr = 0;
            v4l2_buf.m.planes[0].bytesused = v4l2_buf.m.planes[1].bytesused = v4l2_buf.m.planes[2].bytesused = 0;
        }

        if (ctx.copy_timestamp)
        {
          v4l2_buf.flags |= V4L2_BUF_FLAG_TIMESTAMP_COPY;
          ctx.timestamp += ctx.timestampincr;
          v4l2_buf.timestamp.tv_sec = ctx.timestamp / (MICROSECOND_UNIT);
          v4l2_buf.timestamp.tv_usec = ctx.timestamp % (MICROSECOND_UNIT);
        }

        if(ctx.output_memory_type == V4L2_MEMORY_DMABUF || ctx.output_memory_type == V4L2_MEMORY_MMAP)
        {
            for (int j = 0 ; j < MAX_PLANES ; j++)
            {
                ret = NvBufferMemSyncForDevice (buffer->planes[j].fd, j, (void **)&buffer->planes[j].data);
                if (ret < 0)
                {
                    LOGE("Error while NvBufferMemSyncForDevice at output plane for V4L2_MEMORY_DMABUF\n");
                    abort(&ctx);
                    goto cleanup;
                }
            }
        }

        if(ctx.output_memory_type == V4L2_MEMORY_DMABUF)
        {
            for (int j = 0 ; j < MAX_PLANES ; j++)
            {
                v4l2_buf.m.planes[j].bytesused = buffer->planes[j].bytesused;
            }
        }
        ret = ctx.enc->output_plane.qBuffer(v4l2_buf, NULL);
        if (ret < 0)
        {
            LOGE("Error while queueing buffer at output plane\n");
            abort(&ctx);
            goto cleanup;
        }

        if (v4l2_buf.m.planes[0].bytesused == 0)
        {
            LOGE("File read complete.\n");
            eos = true;
            break;
        }
        ctx.input_frames_queued_count++;
    }

    if (ctx.blocking_mode)
    {
        // Wait till capture plane DQ Thread finishes
        // i.e. all the capture plane buffers are dequeued
        if (encoder_proc_blocking(ctx, eos) != 0)
        {
            LOGI("encoder_proc_blocking end, cleanup!\n");
            goto cleanup;
        }
        ctx.enc->capture_plane.waitForDQThread(-1);
    }


cleanup:

    if (NULL != ctx.enc && ctx.enc->isInError())
    {
        LOGE("Encoder is in error\n");
        error = 1;
    }
    if (ctx.got_error)
    {
        error = 1;
    }

    if(NULL != ctx.enc && ctx.output_memory_type == V4L2_MEMORY_DMABUF)
    {
        for (uint32_t i = 0; i < ctx.enc->output_plane.getNumBuffers(); i++)
        {
            ret = ctx.enc->output_plane.unmapOutputBuffers(i, ctx.output_plane_fd[i]);
            if (ret < 0)
            {
                LOGE("Error while unmapping buffer at output plane\n");
                goto cleanup;
            }
            LOGD("begin to NvBufferDestroy!\n");
            ret = NvBufferDestroy(ctx.output_plane_fd[i]);
            if(ret < 0)
            {
                LOGE("Failed to Destroy NvBuffer\n");
                return NULL;
            }
            LOGD("NvBufferDestroy end!\n");
        }
    }
    if (NULL != ctx.enc)
    {
        LOGD("before delete ctx.enc!\n");
        delete ctx.enc;
        LOGD("after delete ctx.enc!\n");
    }

    if (!ctx.blocking_mode)
    {
        LOGD("before sem_destroy(&ctx.pollthread_sema)!\n");
        sem_destroy(&ctx.pollthread_sema);
        sem_destroy(&ctx.encoderthread_sema);
        LOGD("after sem_destroy(&ctx.pollthread_sema)!\n");
    }

    if (error)
    {
        LOGE("App run failed\n");
    }
    else
    {
        LOGD("App run was successful\n");
    }
    return NULL;
}
